/*
 * Software Suspend.
 *
 * Module level definitions.
 *
 */

#ifndef SWSUSP_COMMON_H
#define SWSUSP_COMMON_H

#include <linux/version.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/locks.h>
#include <linux/swapctl.h>
#if defined(CONFIG_FBCON_SPLASHSCREEN)
#include <linux/fb.h>
#include <video/fbcon.h>
#else
#include <linux/console_struct.h>
#endif
#include <linux/suspend.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/file.h>
#include <linux/utsname.h>
#include <linux/init.h>
#include <linux/compiler.h>
#include <linux/delay.h>
#include <linux/reboot.h>
#include <linux/vt_kern.h>
#include <linux/bitops.h>
#include <linux/interrupt.h>
#include <linux/kbd_kern.h>
#include <linux/keyboard.h>
#include <linux/spinlock.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/swap.h>
#include <linux/pm.h>
#include <linux/raid/md.h>
#include <linux/selection.h>
#include <linux/console.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/suspend-debug.h>
#include <linux/zlib.h>

#include <asm/uaccess.h>
#include <asm/mmu_context.h>
#include <asm/pgtable.h>
#include <asm/io.h>
#include <asm/mtrr.h>

#define MAX(a,b)	((signed long)(a)>(signed long)(b)?(a):(b))
#define MIN(a,b)	((signed long)(a)<(signed long)(b)?(a):(b))

struct rangechain {
	struct range * first;
	struct range * last;
	int size; /* size of the range ie sum (max-min+1) */
	int allocs;
	int frees;
	int debug;
	int timesusedoptimisation;
	char * name;
	struct range * lastaccessed, *prevtolastaccessed, *prevtoprev;
};

/*
 * We rely on ranges not fitting evenly into a page.
 * The last four bytes are used to store the number
 * of the page, to make saving & reloading pages simpler.
 */
struct range {
	unsigned long minimum;
	unsigned long maximum;
	struct range * next;
};

/* page backup entry */
typedef struct pbe {
	struct page * origaddress;	/* Original address of page */
	struct page * address;		/* Address of copy of page */
	swp_entry_t swap_address;
	kdev_t swap_bdev;		/* Physical location of swap entry */
	unsigned long block[PAGE_SIZE / 512 ];
	struct range * currentorigrange;
	struct range * currentdestrange;
	struct range * currentswaprange;
	struct range * currentblockrange;

	struct pagedir * pagedir;
	int blocksperpage;
	struct io_info * io_info;
} suspend_pbe_t;

typedef struct pagedir {
	int pagedir_num;
	int pageset_size;
	int lastpageset_size;
	struct rangechain origranges;
	struct rangechain destranges;
	/* NB:
	 * pagedir1 & 2 swaprange chains are actually one big chain, stored in
	 * pagedir2's data. Pagedir2 swap comes before pagedir1 swap (save
	 * order), and the first swap entry for pagedir1 to use is set when
	 * pagedir2 is written (when we know how much swap it used). Since
	 * this first entry is almost certainly not at the start of a range,
	 * the firstoffset variable below tells us where to start in the range.
	 * All of this means we don't have to worry about getting different
	 * compression ratios for the kernel and cache. We can simply allocate
	 * one pool of swap (size determined using expected compression ratio)
	 * and use it without worrying whether one pageset compresses better
	 * and the other worse (this is what happens). As long as the user gets
	 * the expected compression right, it will work.
	 */
	struct rangechain swapranges;
	unsigned long firstswapoffset;
	struct range * firstblockrange;
	unsigned long firstblockoffset;

	struct rangechain allocdranges;

	/* 
	 * Ranges of blocks that we are using within each swapfile.
	 * These are the blocks that match the swap entries stored
	 * in the swapranges chain.
	 */
	struct rangechain block_chain[MAX_SWAPFILES];
} suspend_pagedir_t;

struct pbelink {
	char dummy[sizeof(struct pbe) - sizeof(struct pbe *)];
	struct pbe * next;
};

#define SWAP_FILENAME_MAXLENGTH	32

struct suspend_header {
	__u32 version_code;
	unsigned long num_physpages;
	char machine[65];
	char version[65];
	int num_cpus;
	int page_size;
	unsigned long orig_mem_free;
	unsigned long swapforimage;
	int pages_written;
	int num_range_pages;
	struct pagedir pagedir;
	int pageset_2_size;
	int param0;
	int param1;
	int param2;
	int param3;
	int param4;
	int progress0;
	int progress1;
	int progress2;
	int progress3;
	int progress4;
	kdev_t kdevs[MAX_SWAPFILES];
	char blocksize[MAX_SWAPFILES];
	struct rangechain block_chain[MAX_SWAPFILES];
};

extern struct tq_struct suspend_tq;

struct io_info {
	struct buffer_head * bh;
	struct page * cache_page;
	struct page * buffer_page;
	char flags;
	pte_t * pte;
	int pageprot;
	struct page * data_page;
	swp_entry_t swap_address;
	kdev_t dev;
	int zones[PAGE_SIZE/512];
	int zones_used;
	int block_size;
	struct io_info * next;
};

/* mm/vmscan.c */
extern int shrink_mem(void);

#ifndef SWSUSP_MAIN_C
//extern unsigned char software_suspend_disabled;
extern unsigned int suspend_task;
extern unsigned long swsusp_debug_state;
extern char swapfilename[];
extern struct sysinfo swapinfo;
extern unsigned short root_swap;
extern kdev_t swapdevs[MAX_SWAPFILES];
extern spinlock_t suspend_irq_lock __nosavedata;
extern unsigned long swsuspirqflags __nosavedata;

extern int max_async_ios;
extern int image_size_limit;
extern int __nosavedata pages_written;
extern int expected_compression;

extern int currentbeep;
extern int orig_loglevel;
extern int orig_show_page_allocs;

extern int now_resuming;
extern struct pagedir pagedir1, pagedir2;
extern unsigned long orig_mem_free;
#endif

#ifndef SWSUSP_MEMORY_POOL_C
extern int swsusp_pool_level;
#endif

struct swaplink {
	char dummy[PAGE_SIZE - sizeof(swp_entry_t)];
	swp_entry_t next;
};

union diskpage {
	union swap_header swh;	/* swh.magic is the only member used */
	struct swaplink link;
	struct suspend_header sh;
};

union p_diskpage {
	union diskpage *pointer;
	char *ptr;
        unsigned long address;
};

typedef struct pageset_sizes_result {
	int size1; /* Can't be unsigned - breaks MAX function */
	int size2;
	int size2low;
	int needmorespace;
} pageset_sizes_t;

/* */
extern int C_A_D;

/*
 * Poll the swsusp state every second
 */
#define SWSUSP_CHECK_TIMEOUT	(HZ)
#define SWSUSP_COMMIT_TIMEOUT	(HZ)

#if defined(CONFIG_VT) && defined(CONFIG_VT_CONSOLE)
#define DEFAULT_SUSPEND_CONSOLE	(MAX_NR_CONSOLES-1)
#endif

#define TIMEOUT	(6 * HZ)	/* Timeout for stopping processes */
#define __ADDRESS(x)  ((unsigned long) phys_to_virt(x))
#define ADDRESS(x) __ADDRESS((x) << PAGE_SHIFT)

#define MB(x) ((x) >> (20 - PAGE_SHIFT))

#define MARK_SWAP_SUSPEND 0
#define MARK_SWAP_RECOVER 1
#define MARK_SWAP_RESUME 2

/* References to section boundaries */
extern char _text, _etext, _edata, __bss_start, _end;
extern char __nosave_begin, __nosave_end;

extern int console_loglevel;
extern inline void signal_wake_up(struct task_struct *t);
asmlinkage long sys_sched_yield(void);
asmlinkage void sys_sync(void);	/* it's really int */
extern void ide_disk_unsuspend(int);
extern void ide_disk_suspend(void);
extern kdev_t name_to_kdev_t(char *line) __init;
extern void vt_console_print(struct console *co, const char * b, unsigned count);
extern void free_suspend_cache_page(struct page * page);
extern inline pte_t *lookup_address(unsigned long address);
extern int try_to_free_pages_zone_swsusp(zone_t *classzone, unsigned int gfp_mask, int priority, int nr_pages);
extern int try_to_free_pages_swsusp(unsigned int gfp_mask, int amount_needed);
extern char log_buf[];

#ifdef CONFIG_PREEMPT
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
#define swsusp_spin_lock_irq(lock) { \
	int count = current->preempt_count; \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinlock %p. %d->%d.\n", lock, count, count+1); \
	spin_lock_irq(lock); \
}

#define swsusp_spin_lock_irqsave(lock, flags) { \
	int count = current->preempt_count; \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinlock IRQSave %p. %d->%d.\n", lock, count, count+1); \
	spin_lock_irqsave(lock, flags); \
}

#define swsusp_spin_unlock_irq(lock) { \
	int count = current->preempt_count; \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinunlock %p. %d->%d.\n", lock, count, count-1); \
	spin_unlock_irq(lock); \
}

#define swsusp_spin_unlock_irqrestore(lock, flags) { \
	int count = current->preempt_count; \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinunlock IRQRestore %p. %d->%d\n", lock, count, count-1); \
	spin_unlock_irqrestore(lock, flags); \
}
#else /* No debug */
#define swsusp_spin_lock_irq(lock) spin_lock_irq(lock)
#define swsusp_spin_lock_irqsave(lock, flags) spin_lock_irqsave(lock, flags)
#define swsusp_spin_unlock_irq(lock) spin_unlock_irq(lock)
#define swsusp_spin_unlock_irqrestore(lock, flags) spin_unlock_irqrestore(lock, flags)
#endif
#else /* No preempt */
#define swsusp_spin_lock_irq(lock) { \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinlock %p.\n", lock); \
	spin_lock_irq(lock); \
}

#define swsusp_spin_lock_irqsave(lock, flags) { \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinlock IRQSave %p.\n", lock); \
	spin_lock_irqsave(lock, flags); \
}

#define swsusp_spin_unlock_irq(lock) { \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinunlock %p.\n", lock); \
	spin_unlock_irq(lock); \
}

#define swsusp_spin_unlock_irqrestore(lock, flags) { \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinunlock IRQRestore %p.\n", lock); \
	spin_unlock_irqrestore(lock, flags); \
}
#endif

/* Variables to be preserved over suspend */
extern char resume_file[256];	/* For resume= kernel option */
/* Local variables that should not be affected by save */
#define pageset1_size (pagedir1.pageset_size)
#define pageset2_size (pagedir2.pageset_size)

#define BITS_PER_PAGE (PAGE_SIZE * 8)
#define PAGES_PER_BITMAP ((max_mapnr + BITS_PER_PAGE - 1) / BITS_PER_PAGE)
#define BITMAP_ORDER (get_bitmask_order((PAGES_PER_BITMAP) - 1))

/*
 * XXX: We try to keep some more pages free so that I/O operations succeed
 * without paging. Might this be more?
 */
#define MIN_FREE_RAM (max_mapnr >> 7)

void prepare_status(int printalways, int clearbar, const char *fmt, ...);
void abort_suspend(const char *fmt, ...);

/* Exported variables */

int read_secondary_pagedir(int);
int read_primary_suspend_image(int noresume);
void thaw_processes(void);

extern int nr_unused_buffer_heads;

/* ------ memory_pool.c ------ */

void fill_swsusp_memory_pool(void);
void empty_swsusp_memory_pool(void);

/* ------ low_level_io.c ------ */

extern int get_phys_params(struct pagedir * pagedir, swp_entry_t entry);

/* ------ io.c ------ */
int write_pageset(struct pagedir * pagedir, int whichtowrite);
int read_pageset(struct pagedir * pagedir, int whichtoread, int overwrittenpagesonly);
int write_pagedir_and_suspend_header(void);
void mark_swapfiles(struct pbe * headerpbe, int mode);
void sync_swap_partitions(void); 
int noresume_fix_signature(union p_diskpage cur);
void do_suspend_sync(void);
void read_swapfiles(void);
void finish_all_io(void);
int bdev_page_io(int rw, kdev_t bdev, long pos, struct page * page);
extern kdev_t resume_device;
extern int resume_firstblock;
extern int resume_firstblocksize;
int do_swsusp_io(int rw, struct pbe * pbe, int asyncio);

/* ------ ui.c ------ */
void check_shift_keys(int pause, char * message);
int update_status(unsigned long value, unsigned long maximum, const char *fmt, ...);
#if defined(CONFIG_FBCON_SPLASHSCREEN)
extern int fb_splash_set_progress(int consolenr, int value, int maximum, int reset);
extern int fb_splash_progress[5]; /* Needed while progress bar posn/size set via proc interface */
int splash_is_on(int consolenr);
#define SPLASH_IS_ON(consolenr) (splash_is_on(consolenr))
#else
#define SPLASH_IS_ON(consolenr) (0)
#define fb_splash_set_progress(a, b, c, d) do { } while (0)
#endif
int prepare_suspend_console(void);
void restore_console(void);

#define PBEPAGE_START(pbe) (((unsigned long) (pbe)) & PAGE_MASK)
void get_next_pbe(struct pbe * pbe);
void get_first_pbe(struct pbe * pbe, struct pagedir * pagedir);

#define REAL_MAX_ASYNC ((max_async_ios ? max_async_ios : 32))
#define io_pages (REAL_MAX_ASYNC * sizeof(struct io_info) / PAGE_SIZE + 1)
/* Spare is space for extra range pages allocated in save_image */
#define SPARE 10

#ifdef CONFIG_SOFTWARE_SUSPEND_COMPRESSION
#define EXPECTED_RATIO (TEST_ACTION_STATE(SUSPEND_NO_COMPRESSION) ? \
	100 : (100 - expected_compression))
#else
#define EXPECTED_RATIO 100
#endif
#define SWAP_NEEDED(include_spare) ((pageset1_size + pageset2_size + 1 + \
	num_range_pages + (include_spare ? SPARE : 0)) * EXPECTED_RATIO / 100)
#define RAM_TO_SUSPEND(include_spare) (1 + \
	MAX((pageset1_size - pageset2_sizelow), 0) + MIN_FREE_RAM + \
	REAL_MAX_ASYNC + io_pages + (include_spare ? SPARE : 0))

/* --------- Range chains ----------- */

#define RANGES_PER_PAGE (PAGE_SIZE / (sizeof(struct range)))
#define RANGE_PAGES_NEEDED(x) (((x)+RANGES_PER_PAGE-1)/RANGES_PER_PAGE)
#define RANGEPAGELINK(x) ((unsigned long *) ((((unsigned long) x) & PAGE_MASK) + PAGE_SIZE - sizeof(unsigned long)))

#define range_for_each(rangechain, rangepointer, value) \
if ((rangechain)->first) \
	for ((rangepointer) = (rangechain)->first, (value) = (rangepointer)->minimum; \
	     ((rangepointer) && ((rangepointer)->next || (value) <= (rangepointer)->maximum)); \
	     (((value) == (rangepointer)->maximum) ? \
		((rangepointer) = (rangepointer)->next, (value) = ((rangepointer) ? (rangepointer)->minimum : 0)) : \
	     	(value)++))

/*
 * When using compression and expected_compression > 0,
 * we allocate less swap entries, so GET_RANGE_NEXT can
 * validly run out of data to return.
 */
#define GET_RANGE_NEXT(currentrange, currentval) \
{ \
	if (currentrange) { \
		if ((currentval) == (currentrange)->maximum) { \
			if ((currentrange)->next) { \
				(currentrange) = (currentrange)->next; \
				(currentval) = (currentrange)->minimum; \
			} else { \
				(currentrange) = NULL; \
				(currentval) = 0; \
			} \
		} else \
			currentval++; \
	} \
}

struct range_entry {
	unsigned long thisvalue;
	struct range * rangeptr;
};

extern int max_ranges_used;
extern int num_range_pages;
int add_range_pages(int number_requested);
int add_to_range_chain(struct rangechain * chain, unsigned long value);
void put_range_chain(struct rangechain * chain);
void print_chain(int debuglevel, struct rangechain * chain, int printasswap);
void set_chain_names(struct pagedir * p);
int free_ranges(void);
void init_range_entry(struct range_entry * range_entry, struct rangechain rangechain);
int append_range_to_range_chain(struct rangechain * chain, unsigned long minimum, unsigned long maximum);
void relativise_ranges(void);
void relativise_chain(struct rangechain * chain);
void absolutise_ranges(unsigned long * range_pages);
void absolutise_chain(struct rangechain * chain, unsigned long * range_pages);
unsigned long * get_rangepages_list(void);
void put_rangepages_list(unsigned long * range_pages);
int relocate_rangepages(unsigned long * rangepages);

extern struct range * first_range_page, * last_range_page;

/* We are putting offset in the low bits so consecutive swap entries make consecutive range values */
#define swap_entry_to_range_val(swp_entry) ((swp_entry.val >> 8) | ((swp_entry.val & 0x3f) << 24))
#define range_val_to_swap_entry(val) (swp_entry_t) { ((val >> 24) | ((val & 0xffffff) << 8)) }
#define RANGE_RELATIVE(x) (struct range *) ((((unsigned long) x) & (PAGE_SIZE - 1)) | \
				((*RANGEPAGELINK(x) & (PAGE_SIZE - 1)) << PAGE_SHIFT))
#define RANGE_ABSOLUTE(entry, list) (struct range *) ((((unsigned long) (entry)) & (PAGE_SIZE - 1)) | \
				list[((unsigned long) (entry)) >> PAGE_SHIFT])

extern unsigned long * inusemap;
extern unsigned long * pageset2map;
extern unsigned long * nosavemap;

#define PAGENUMBER(page) (page-mem_map)
#define PAGEINDEX(page) ((PAGENUMBER(page))/(8*sizeof(unsigned long)))
#define PAGEBIT(page) ((int) ((PAGENUMBER(page))%(8 * sizeof(unsigned long))))

/* 
 * freepagesmap is used in two ways: 
 * - During suspend, to tag pages which are not used (to speed up count_data_pages);
 * - During resume, to tag pages which are is pagedir1. This does not tag pagedir2
 *   pages, so !== first use.
 */
#define PageInUse(page)		test_bit(PAGEBIT(page), &inusemap[PAGEINDEX(page)])
#define SetPageInUse(page)	set_bit(PAGEBIT(page), &inusemap[PAGEINDEX(page)])
#define ClearPageInUse(page)	clear_bit(PAGEBIT(page), &inusemap[PAGEINDEX(page)])

#define PagePageset2(page)	test_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])
#define SetPagePageset2(page)	set_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])
#define TestAndSetPagePageset2(page)	test_and_set_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])
#define ClearPagePageset2(page)	clear_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])

/* We don't allocate Nosavemap when reloading the kernel, and 
 * don't care about whether Nosave is set or clear */
#define PageNosave(page)	(nosavemap ? test_bit(PAGEBIT(page), &nosavemap[PAGEINDEX(page)]) : 0)
#define SetPageNosave(page)	(nosavemap ? set_bit(PAGEBIT(page), &nosavemap[PAGEINDEX(page)]) : 0)
//extern void SetPageNosave(struct page * page);
#define ClearPageNosave(page)	(nosavemap ? clear_bit(PAGEBIT(page), &nosavemap[PAGEINDEX(page)]) : 0)
#endif
